﻿using System;
using System.Collections.Generic;
using System.Runtime.Serialization;
using System.Windows.Forms;
using System.Threading;
using System.Reflection;
using System.IO;
using System.Linq;
using System.Xml.Linq;
using Microsoft.Win32;

namespace eWalletDataLiberator
{
	static class Program
	{
		private static FieldInfo eWalletMainFormField;
		private static FieldInfo eWalletMainFormWalletField;

		private static string sEWalletPath = null;

		private static bool sEWalletExited = false;
		
		[STAThread]
		static void Main()
		{
			AppDomain.CurrentDomain.AssemblyResolve += OnAssemblyResolve;

			if (FindEWallet())
			{
				if (FindObfuscatedMethods())
				{
					LiberateData();
				}
				else
				{
					MessageBox.Show("Could not guess compatibility with this version of eWallet. Please see Readme.txt", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				}
			}
			else
			{
				MessageBox.Show("Could not find eWallet Install Folder. Please try reinstalling it.", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
			
			AppDomain.CurrentDomain.AssemblyResolve -= OnAssemblyResolve;
		}

		private static bool FindEWallet()
		{
			sEWalletPath = (string)Registry.GetValue(@"HKEY_CURRENT_USER\Software\Ilium Software\eWallet", "InstallDir", null);
			if (String.IsNullOrEmpty(sEWalletPath))
			{
				return false;
			}

			return true;
		}

		private static bool FindObfuscatedMethods()
		{
			var eWalletApplicationType = typeof(eWallet.MainForm).Assembly.EntryPoint.DeclaringType;
			// Find MainForm field
			eWalletMainFormField = eWalletApplicationType.GetFields(BindingFlags.Static | BindingFlags.Public).SingleOrDefault(field => field.FieldType == typeof(eWallet.MainForm));
			if (eWalletMainFormField == null) return false;

			// Find Wallet Field
			eWalletMainFormWalletField = typeof(eWallet.MainForm).GetFields(BindingFlags.Instance | BindingFlags.NonPublic).SingleOrDefault(field => field.FieldType == typeof(DBWalletDotNet.Wallet));
			if (eWalletMainFormField == null) return false;

			return true;
		}

		private static void LiberateData()
		{
			Thread waitForEWallet = new Thread(WaitForEWallet);
			waitForEWallet.SetApartmentState(ApartmentState.STA);
			waitForEWallet.Start();

			NeuterSparkleDotNet();

			//Execute eWallet (this is synchronous)
			typeof(eWallet.MainForm).Assembly.EntryPoint.Invoke(null, new object[] { new string[0] });
			sEWalletExited = true;
		}

		private static void NeuterSparkleDotNet()
		{
			Assembly sparkleDotNet;
			try
			{
				sparkleDotNet = Assembly.Load("Updater");
				if (sparkleDotNet == null) return;		
			}
			catch (Exception)
			{
				return;
			}
			
			var kBundleType = sparkleDotNet.GetType("SparkleDotNET.KNBundle");
			if (kBundleType == null) return;
			
			var kBundleCacheField = kBundleType.GetField("bundleCache", BindingFlags.NonPublic | BindingFlags.Static);
			if (kBundleCacheField == null) return;
			
			var kBundleCache = kBundleCacheField.GetValue(null) as System.Collections.IDictionary;
			if (kBundleCache == null) return;

			var blankBundle = FormatterServices.GetUninitializedObject(kBundleType);
			if (blankBundle == null) return;

			kBundleCache.Add(new FileInfo(Assembly.GetEntryAssembly().Location).DirectoryName, blankBundle);

			var suupdaterType = sparkleDotNet.GetType("SparkleDotNET.SUUpdater");
			if (suupdaterType == null) return;

			var sharedUpdatersField = suupdaterType.GetField("sharedUpdaters", BindingFlags.NonPublic | BindingFlags.Static);
			if (sharedUpdatersField == null) return;

			var sharedUpdaters = sharedUpdatersField.GetValue(null) as System.Collections.IDictionary;
			if (sharedUpdaters == null) return;

			sharedUpdaters.Add(blankBundle, null);
		}

		private static Assembly OnAssemblyResolve(object sender, ResolveEventArgs args)
		{
			var splitPoint = args.Name.IndexOf(',');
			var localName = splitPoint == -1 ? args.Name : args.Name.Substring(0, splitPoint);
			
			var eWalletFilePath = Path.Combine(sEWalletPath, localName) + (localName.Equals("ewallet", StringComparison.InvariantCultureIgnoreCase) ? ".exe" : ".dll");
			if (File.Exists(eWalletFilePath))
			{
				return Assembly.LoadFile(eWalletFilePath);
			}

			return null;
		}

		private static void WaitForEWallet()
		{
			//Wait for main form to be available
			eWallet.MainForm form = null;
			int timeout = 0;
			while ((form = (eWallet.MainForm)eWalletMainFormField.GetValue(null)) == null)
			{
				Thread.Sleep(500);
				if (sEWalletExited)
				{
					MessageBox.Show("eWallet is already running. Please exit it before running " + Application.ProductName, Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
					return;
				}
				if (++timeout > 40) //If over 20s, timeout.
				{
					MessageBox.Show("Could not connect to eWallet. If it is running, please exit it before running " + Application.ProductName, Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
					return;
				}
			}

			//Main form is now available, wait for wallet to be available
			if (MessageBox.Show("Go to eWallet, open the wallet or template to extract the data from (including entering the password), then click OK", Application.ProductName, MessageBoxButtons.OKCancel) != DialogResult.OK)
			{
				//Exit.
				CloseEWallet(form);
				return;
			}

			if (form.IsDisposed)
			{
				MessageBox.Show("eWallet connection lost. Please do not exit eWallet until prompted. " + Application.ProductName + " will now exit.", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}

			DBWalletDotNet.Wallet wallet = null;
			while ((wallet = (DBWalletDotNet.Wallet)eWalletMainFormWalletField.GetValue(form)) == null ||
				   !wallet.IsOpen)
			{
				if (MessageBox.Show("No open wallet detected. Please open the wallet or template to extract the data from (including entering the password), then click OK", Application.ProductName, MessageBoxButtons.OKCancel) != DialogResult.OK)
				{
					//Exit.
					CloseEWallet(form);
					return;
				}
			};

			var saveFileDialog = new SaveFileDialog()
									{
										Title = String.Format("Save extracted data from {0} ({1} cards)", Path.GetFileName(wallet.Path), wallet.Cards.AllCards.Count),
										Filter = "XML files (*.xml)|*.xml|All Files (*.*)|*.*",
										DefaultExt = "xml"
									};

			StreamWriter outputFile;
				
			while (true) //Until a valid file breaks out
			{
				if (saveFileDialog.ShowDialog() == DialogResult.Cancel)
				{
					//Exit.
					CloseEWallet(form);
					return;
				}
				//Try to create the file
				try
				{
					outputFile = File.CreateText(saveFileDialog.FileName);
					//File seems valid, break out
					break;
				}
				catch (Exception ex)
				{
					if (MessageBox.Show("Could not create file: " + saveFileDialog.FileName + "\n\n" + ex.Message, Application.ProductName, MessageBoxButtons.RetryCancel) == DialogResult.Cancel)
					{
						//Exit.
						CloseEWallet(form);
						return;
					}
				}
			}

			if (form.IsDisposed)
			{
				MessageBox.Show("eWallet connection lost. Please do not exit eWallet until prompted. " + Application.ProductName + " will now exit.", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}

			var xWallet = new XElement("Wallet");

			//Get categories
			var categoryLookup = new Dictionary<long, XElement>();
			foreach (long categoryId in wallet.Categories.GetChildCategoryIds(0))
			{
				xWallet.Add(CreateXCategory(wallet, categoryLookup, categoryId));
			}

			foreach (long cardId in wallet.Cards.AllCards)
			{
				//Construct a card for reading that data
				var card = new DBWalletDotNet.Card(wallet, cardId);

				XElement xCard;
				if (card.IsShortcut)
				{
					xCard = new XElement("Shortcut");

					xCard.Add(new XAttribute("id", cardId));
					xCard.Add(new XAttribute("name", card.Name));

					xCard.Add(new XAttribute("target", card.ShortcutSourceCardId));
				}
				else
				{
					//Write data out to XML file
					xCard = new XElement("Card");

					xCard.Add(new XAttribute("id", cardId));
					xCard.Add(new XAttribute("name", card.Name));

					xCard.Add(new XAttribute("kind", card.Kind));
					if (card.TemplateName != null)
					{
						xCard.Add(new XAttribute("templateName", card.TemplateName));
					}

					if (card.BackgroundColor.HasValue)
					{
						xCard.Add(new XAttribute("backgroundColor", System.Drawing.ColorTranslator.ToHtml(card.BackgroundColor.Value)));
					}
					if (card.TextColor.HasValue)
					{
						xCard.Add(new XAttribute("textColor", System.Drawing.ColorTranslator.ToHtml(card.TextColor.Value)));
					}

					xCard.Add(new XAttribute("showGradient", card.Gradient == DBWalletDotNet.ShowHide.Show));
					xCard.Add(new XAttribute("showIcon", card.Icon == DBWalletDotNet.ShowHide.Show));
					xCard.Add(new XAttribute("showBorder", card.Border == DBWalletDotNet.ShowHide.Show));
					xCard.Add(new XAttribute("showScrollBars", card.ScrollBars == DBWalletDotNet.ShowHide.Show));

					xCard.Add(new XAttribute("viewKind", card.ViewKind));

					xCard.Add(new XAttribute("cornerShape", card.CornerShape));
					if (card.FontSize.HasValue)
					{
						xCard.Add(new XAttribute("fontSize", card.FontSize.Value));
					}
					xCard.Add(new XAttribute("imagePosition", card.ImagePosition));


					xCard.Add(new XAttribute("iconKind", card.IconKind));
					xCard.Add(new XAttribute("standardIcon", card.StandardIcon));
					if (card.CustomIcon != null)
					{
						xCard.Add(new XAttribute("customIcon", card.CustomIcon));
					}

					xCard.Add(new XElement("Face", CreateXFields(card, card.AllFaceFields)));
					xCard.Add(new XElement("Detail", CreateXFields(card, card.AllDetailsFields)));

					xCard.Add(new XElement("Notes", card.Notes));

					
					AddAttachment(xCard, card, DBWalletDotNet.FileAttachmentSlot.Graphics, "Image");
					AddAttachment(xCard, card, DBWalletDotNet.FileAttachmentSlot.Generic, "Attachment");

				}
				var xCategory = categoryLookup[card.CategoryId];
				xCategory.Add(xCard);

			}

			xWallet.Save(outputFile);

			CloseEWallet(form);
			MessageBox.Show("Data extraction complete, eWallet has been closed.", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Information);
		}

		private static void CloseEWallet(Form form)
		{
			if (!form.IsDisposed)
			{
				form.Invoke(new Action(form.Close));
			}
		}

		private static void AddAttachment(XElement xCard, DBWalletDotNet.Card card, DBWalletDotNet.FileAttachmentSlot slot, string name)
		{
			var kind = card.GetAttachedFileType(slot);
			if (kind != DBWalletDotNet.FileAttachmentKind.None)
			{
				var attachment = new XElement(name, new XAttribute("path", card.GetAttachedPath(slot)));
				if (kind == DBWalletDotNet.FileAttachmentKind.Embedded)
				{
					attachment.Add(System.Convert.ToBase64String(card.GetEmbeddedFileData(slot)));
				}
				xCard.Add(attachment);
			}
		}

		private static IEnumerable<XElement> CreateXFields(DBWalletDotNet.Card card, IEnumerable<int> fieldIDs)
		{
			foreach (var fieldID in fieldIDs)
			{
				var xField = new XElement("Field");
				xField.Add(new XAttribute("label", card.GetFieldLabel(fieldID)));
				var kind = card.GetFieldKind(fieldID);
				xField.Add(new XAttribute("kind", kind));
				
				var value = card.GetFieldValue(fieldID);
				if (!String.IsNullOrEmpty(value))
				{
					if (kind == DBWalletDotNet.FieldKind.Date)
					{
						xField.Add(DBWalletDotNet.Card.DateFromFieldValue(value));
					}
					else
					{
						xField.Add(value);
					}
				}

				yield return xField;
			}
		}

		private static XElement CreateXCategory(DBWalletDotNet.Wallet wallet, Dictionary<long, XElement> categoryLookup, long categoryId)
		{
			var category = new DBWalletDotNet.Category(wallet, categoryId);
			var xCategory = new XElement("Category");
			xCategory.Add(new XAttribute("id", categoryId));
			xCategory.Add(new XAttribute("name", category.Name));
			xCategory.Add(new XAttribute("defaultCardKind", category.DefaultCardKind));
			xCategory.Add(new XAttribute("defaultCardTemplate", category.DefaultCardTemplate));
			xCategory.Add(new XAttribute("iconKind", category.IconKind));
			xCategory.Add(new XAttribute("standardIcon", category.StandardIcon));
			if (category.CustomIcon != null)
			{
				xCategory.Add(new XAttribute("customIcon", category.CustomIcon));
			}

			categoryLookup.Add(categoryId, xCategory);

			//Write out child categories
			foreach (long childId in wallet.Categories.GetChildCategoryIds(categoryId))
			{
				xCategory.Add(CreateXCategory(wallet, categoryLookup, childId));
			}

			return xCategory;
		}
	}
}
